#include "../libcryst/cryst.h"
#include "../libgnuplot/gplot.h"
#include "../libbild/bild.h"

#include<values.h>
#include <ctype.h>
#include <stdlib.h>
#include<glib/gmain.h>
#include<glib/glist.h>

#ifndef IPDS_H
#define IPDS_H

#define GROESSE 800

#define INITIAL 1
#define MEHRLING 2
#define PHASE 3

#define Null 0
#define BG 1

#define MAX_TRIG 0
#define MITTL_TRIG 1


#define PI2 6.2831853
#ifndef max
#define max(a,b) ( (a)>(b)? (a):(b) )
#endif
#ifndef min
#define min(a,b) ( (a)<(b)? (a):(b) )
#endif

typedef struct {
  double phi;
  int ij;
} geo_arr_t;

typedef struct g_list {
  int r;
  double phi;
  int ij;
  struct g_list *next;
} geo_list_t;

typedef struct dlist {
  double decay;
  int platte;
  int run;
  struct dlist *next;
} decay_list_t;


typedef struct {
  double x_pos;
  double y_pos;
  int integral;
  int untergrund;
} print_peak_t;


typedef struct {
  int do_view;
  int do_view_line;
  int projektions_richtung;
  int projektions_typ;
  int anzahl_proj;
  int background;
  int skalierung;
  int pos_calc[4];
  int run_nr;
  peak_t peak[1];
  double osc_angle[2];
  double zoom;
} view_t;

typedef struct {
  int p_x;
  int p_y;
  int p_z;
  int direction;
  int typ;
  double center[3];
  double center_proj[2];
  double R[9];
  double scale;
  double size;
  //  double acceptance;
  double skal_prod_wert;
} rec_map_t;

typedef struct {
  double phi;
  double chi;
  double psi;
  double indexed;
  double ub[9];
} loes_t;

typedef struct {
  int index;
  int index_new;
  unsigned int wert;
} value_data_t;

typedef struct {
  int pos1[2];
  int pos2[2];
  int peaks[2];
  int integral;
  double steigung;
  double abschnitt;
  double mittel;
  double normale[3];
} gerade_t; 

typedef struct {
  gerade_t *gerade; 
  int anzahl;
} geraden_t;

#define ROTATION 1
#define PRIMAER 2

typedef struct {
  int grid;
  int accept;
  int verfeinerungs_typ;
  int direction[3];
  double periode;
  double max_tan_theta;
  double incr;
  double incr2;
  double delta_w;
  double faktor[2];
  double abweichung[3];
  double normal[9];
  geraden_t geraden[1];
} gnomonisch_t;

typedef struct {
  unsigned int *Daten;
  int x;
  int y;
  int origin[2];
  double scale[2];
} projection_t;

typedef struct {
  double line[2][602];
  float fline[301];
  int anzahl_projektions_punkte;
  double incr[2];
} line_t;


typedef struct {
  double om[9];
  double G[9];
  double indexed;
  int index_G;
  int anzahl_metrik;
  int anzahl_om;
  int individuum;
  double abweichung[3];
} loesung_t;

typedef struct {
  int anzahl;
  loesung_t *loesung;
} loesungen_t;


typedef struct {
  double diff[3];
  double laenge;
  double pos[2];
  int P[2];
  int anzahl_gleiche_laenge;
  int ok;
  int einsortiert;
} diff_t;

typedef struct {
  int anzahl;
  diff_t *diff;
} diffs_t;

typedef struct {
  double translationen[3];
  double corr;
  int indiziert;
} trans_t; 

typedef struct {
  double A[3];  
  double W[3];  
} metrik_t;

typedef struct {
  int debug;
  int ultra_debug;
  int put_index;
  int ende_festlegen;
  double simul_delta_osc_angle;
  char aktuelle_bild_datei[256];
  char bild_datei[256];
  char letzte_bild_datei[256];
  char fileroot[256];
  char filename[256];
  char bild_dir[256];
  char program_dir[256];
} param_t;

typedef struct {
  unsigned int *Daten;
  int n_r;
  int *n_phi;
  int *ind_phi;
  int gesamtzahl;
  int x;
  int y;
  int xy_origin[2];
  int last_type;
  int overflows;
  double summe;
  double mittelwert;
  double mittl_untergrund;
  double stoe_ii_phi;
  double osc_angle;
  double *delta_phi;
  double scale;
} az_data_t;

#define FREE_FIGURE 0
#define ELLIPSE_FIGURE 1

#define BGFIT 1
#define FIX_BACKGROUND 2

#define MATRIX_FREI 1
#define FIX_MATRIX 2
#define PEAK_SEARCH 3

typedef struct {
  int max_fenster[2];
  //  int view_peaks;
  int do_k_alpha12;
  int background_typ;
  int integrations_typ;
  int figur_typ;
  int full_matrix;
  int min_anzahl_relevant;
  int fix_delta_radius;
  int scale;
  int lattice_typ;
  int refine_orientation;
  int anzahl_reorient;
  int max_anzahl_platten;
  int R_min;
  int R_max;
  double min_L;
  double min_sr_reorient;
  double EMS;
  double fix_delta_phi;
  double min_sr;
  double r_min;
  double phi_min;
  double theta_min;
  double theta_max;
  double bgfit_breite;
  double bgfit_ratio;
  double akzeptance;
} peaksearch_t;


typedef struct {
  int H[3];
  int start_aufnahme;
  int end_aufnahme;
  double osc_angle;
  double x_pos[2];
} valid_peaks_t;

typedef struct {
  int max_anzahl_individuen;
  int max_anzahl_typ;
  int max_anzahl_reflexe;
  int only_one_metrik;
  int check_mehrling;
  int not_indexed;
  int anzahl_grob;
  int refine_metrik;
  int reindex_in_calibrate;
  int refine_metrik_in_calibrate;
  int refine_parameter[4];
  int refine_osc_angle_value;
  int refine_all;
  int d_wert_ordnung;
  int dimensions;
  double min_A;
  double max_A;
  double min_laenge;
  double max_laenge;
  double min_diff_laenge;
  double max_diff_laenge;
  double asymmetrie;
  double toleranz;
  double max_dev;
  double volumen;
  double delta_winkel;
  double min_residuum;
  double min_sr;
  double x_tolerance;
  double intern_schwelle;
  double quasilattice[2];
} index_param_t;


typedef struct {
  int *ij_to_rp;
  int *rp_to_ij;
  int *n_phi;
  double *phi;
  int gesamtzahl;
} trans_table_t;

typedef struct {
  int run;
  int file_nr;
  int ij;
  int anzahl;
  //  double phi;
  int index;
} rec_h_t;

typedef struct h_list {
  int H[3];
  int H_q[6];
  int dim;
  int indiv_nr;
  struct h_list *next;
  //  struct h_list *previous;
} hkl_list_t;

typedef struct p_list {
  int r;
  int ind_phi[2];
  struct p_list *next;
  struct p_list *previous;
} position_list_t;

typedef struct o_list {
  int pos[2];
  int max_integral;
  int integral;
  int untergrund;
  int overflow;
  int boundary;
  int platte;
  int platten[2];
  double osc_angle;
  double stoe_ii_phi;
  int run_nr;
  int run_offset;
  int anzahl_relevant;
  hkl_list_t *hkl_list;
  position_list_t *position_list;
  struct o_list *next;
} object_list_t;

typedef struct pl_peak_list {
  int platte;
  int indiv_nr;
  int H[3];
  hkl_list_t *hkl_list;
  double x_pos[2];
  double phi_calc;
  double stoe_ii_phi;
  struct pl_peak_list *next;
} pl_peak_list_t;

typedef struct {
  unsigned int wert;
  int index;
} value_t;
void print_val(gpointer val, gpointer foo);
gint value_groesser(gconstpointer a,gconstpointer b);

typedef struct l_list {
  double phi;
  double chi;
  double psi;
  double indexed;
  double ub[9];
  struct l_list *next;
  struct l_list *previous;
} loes_list_t;

// files_IPDS.c
position_list_t *read_shadow_file(param_t *param,datacol_t *datacol, az_data_t *az_data,trans_table_t *trans_table);
void get_filename(char *file, char *dir);
void strip_dir_from_file(char *file, char *out);
void print_correction_typ(datacol_t *datacol);
void generate_filename(char *filename, char *fileroot, char *dir, int file_nr,int run_nr, datacol_t *datacol);
int read_mehr_oma_file(param_t *param, datacol_t *datacol);
int save_mehr_ub (param_t *param, datacol_t *datacol);
int read_oma_file(param_t *param, datacol_t *datacol);
void do_append_ub(datacol_t *datacol, param_t *param);
int put_calculated_into_peak (peak_t *peak, calc_peak_list_t *obj, trans_table_t *trans_table, datacol_t *datacol, az_data_t *az_data, index_param_t *index_param, param_t *param, int run_nr);
int put_obj_into_peak (peak_t *peak, object_list_t *obj, trans_table_t *trans_table, datacol_t *datacol, az_data_t *az_data, index_param_t *index_param, param_t *param);
int put_object_into_peak (peak_t *peak, object_t *object, trans_table_t *trans_table, datacol_t *datacol, az_data_t *az_data, index_param_t *index_param, param_t *param);
int put_pl_peak_into_peak( peak_t *peak, double *x_pos, int platte, double stoe_ii_phi, int *H,  datacol_t *datacol, param_t *param);
void put_search_list_into_file (FILE *fp, object_list_t *op);
void put_obj_from_list_into_peaks(object_list_t *obj, peaks_t *peaks, param_t *param, datacol_t *datacol, az_data_t *az_data, trans_table_t *trans_table, index_param_t *index_param);
void origin_correct_to_origin(double *origin_correct,double *origin,int *xy_origin,double scale);
object_list_t *read_file(param_t *param,peaks_t *peaks, peaks_t *peaks_alle,  index_param_t *index_param,datacol_t *datacol, az_data_t *az_data, trans_table_t *trans_table, object_list_t *obj_list, peaksearch_t *peaksearch);
void init_data_ipds(data_t *data,datacol_t *datacol);
int parse_run_nummer (char *filename, param_t *param, datacol_t *datacol);
int parse_nummer (char *filename, param_t *param, datacol_t *datacol);
void anti_correct (double x_pos[2],double correct[8],double dcp);
void correct_cry (double *x_pos,double *correct,double dcp);
object_list_t *read_search_obj (char *filename, object_list_t *obj_list, datacol_t *datacol, param_t *param);
object_list_t *read_obj (char *filename, object_list_t *obj_list, datacol_t *datacol, param_t *param);
int read_cry (char *filename, peaks_t *peaks, int max_anzahl, datacol_t *datacol);
int anzahl_runs (char *fileroot,datacol_t *datacol, param_t *param);
int anzahl_files (char *fileroot, datacol_t *datacol, int run,param_t *param);
int check_for_correct (char *fileroot, datacol_t *datacol, param_t *param);
void read_correct(char *filename, datacol_t *datacol, param_t *param);
FILE *create_cry_header(char *fileroot, datacol_t *datacol);
void put_object_into_file(FILE *fp, object_t object);
void put_list_object_into_file(FILE *fp, object_list_t *op);
void save_om(FILE *fp, double *om, double *delta, double lambda);
void put_object_into_file(FILE *fp, object_t object);
void put_mehrl_refl_into_hkl_file(FILE *fp, peak_t peak, int *H, int indiv_nr, double scale,int *test);
void put_refl_into_hkl_file(FILE *fp, peak_t peak, int *test, double scale);
FILE *create_peaks_header(char *fileroot, datacol_t *datacol, peaksearch_t *peaksearch);
void put_cry_matrix (FILE *fp, double *om);
void init_cry_refl (FILE *fp);
int read_plain_bild_file (char *filein, data_t *data, datacol_t *datacol);
int read_bild_file (char *filein, az_data_t *az_data, trans_table_t *trans_table, datacol_t *datacol);
int read_az_pck (char *filein, az_data_t *az_data, trans_table_t *trans_table, int binning);
int read_az_mar_ccd (char *filein, az_data_t *az_data, datacol_t *datacol, trans_table_t *trans_table);
int read_az_smart (char *filein, az_data_t *az_data, trans_table_t *trans_table, datacol_t *datacol);
int read_az_bnl (char *filein, az_data_t *az_data, trans_table_t *trans_table, datacol_t *datacol);
int read_az_tra (char *filein, az_data_t *az_data, trans_table_t *trans_table);
int read_az_xi (char *filein, az_data_t *az_data, datacol_t *datacol, trans_table_t *trans_table);
void write_vectors (peaks_t *peaks, char *file, int all);
void save_peak_file(peaks_t *peaks_alle, datacol_t *datacol, index_param_t *index_param, param_t *param,char *hklname, peaksearch_t *peaksearch, az_data_t *az_data);
#define SINGLE 1
#define MEHRLING 2
int object_list_save_hkl (object_list_t *obj, datacol_t *datacol, index_param_t *index_param, param_t *param,char *hklname, az_data_t *az_data, trans_table_t *trans_table, peaksearch_t *peaksearch, int typ);
int create_ins_file (param_t *param, datacol_t *datacol);
int gp_save_hkl (peaks_t *peaks_alle, datacol_t *datacol, index_param_t *index_param, param_t *param,char *hklname, az_data_t *az_data, peaksearch_t *peaksearch);
void generate_fileroot (char *fileroot, char *bild_datei, datacol_t *datacol, param_t *param);

// process_refl.c
int compare_rec_h_file(rec_h_t *g1, rec_h_t *g2);
int compare_print_peaks(print_peak_t *g1, print_peak_t *g2);
int compare_peaks_indexed(peak_t *g1, peak_t *g2);
int compare_peaks_not_indexed(peak_t *g1, peak_t *g2);
int compare_peaks_sr(peak_t *g1, peak_t *g2);
int compare_peaks_L(peak_t *g1, peak_t *g2);
int compare_peaks_sort(peak_t *g1, peak_t *g2);
int compare_peaks_laenge(peak_t *g1, peak_t *g2);
int compare_peaks_laenge_2(peak_t *g1, peak_t *g2);
int compare_peaks_osc_angle(peak_t *g1, peak_t *g2);
int compare_peaks_start(peak_t *g1, peak_t *g2);
int inv_compare_peaks_start(peak_t *g1, peak_t *g2);
double calc_lorentz_of_az_pos(datacol_t *datacol, double osc_angle,double stoe_ii_phi, int pos[2],az_data_t *az_data, trans_table_t *trans_table);
int index_az_position(int platte, int *pos,int run_nr, double stoe_ii_phi, int *H, index_param_t *index_param, trans_table_t *trans_table, az_data_t *az_data, datacol_t *datacol,param_t *param);
int select_richtungen(double *UB, int *index,int anzahl_translationen, peak_t *peak);
void linear_combination(peaks_t *peaks);
void remove_indexed_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param, datacol_t *datacol);
void take_indexed_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param, datacol_t *datacol);
void take_peaks_to_index(peaks_t *peaks_alle,peaks_t *peaks,index_param_t *index_param);
void take_shortest_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param);
void take_longest_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param);
void take_first_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param);
void take_random_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param);
void remove_weak_peaks (peaks_t *alle_peaks, peaks_t *peaks, index_param_t *index_param);
void remove_Lorentz_peaks (peaks_t *alle_peaks, peaks_t *peaks, datacol_t *datacol, double min_L);
void sort_peaks_indexed (peaks_t *peaks);
void sort_peaks_not_indexed (peaks_t *peaks);
void sort_peaks_sr (peaks_t *peaks);
void sort_peaks_L (peaks_t *peaks);
void sort_peaks_laenge (peaks_t *peaks);
void sort_peaks_laenge_2 (peaks_t *peaks);
void sort_peaks_sort (peaks_t *peaks);
void fill_data_to_azim(data_t *data, data_t *data_azim);
void init_peak (peak_t *peak);
void copy_peak (peak_t *peak1, peak_t *peak2);
void copy_peaks (peaks_t *peaks, peaks_t *peaks_alle);
int calc_benachbart (peak_t peak1, peak_t peak2, trans_table_t *trans_table);
void remove_reflections (peaks_t *peaks);
void init_reflections (peaks_t *peaks);
void put_reflections_on_grid (peaks_t *peaks,int grid);
int check_in_a_row(peaks_t *peaks,int n);
void scale_refl_to_F (peaks_t *peaks);
void calc_peak_geometry(peaks_t *peaks, datacol_t *datacol, param_t *param);
void add_hkl_to_object(object_t *o, int *H,int ind_nr);
void add_hkl_to_object_list(object_list_t *o, int *H,int ind_nr);
int is_new_hkl(object_t *o, int *H,int ind_nr);
int is_new_hkl_list(object_list_t *o, int *H,int ind_nr);
void scale_object_list(object_list_t *obj, decay_list_t *decay);
int put_hkl_to_object_list_from_index(object_list_t *obj, datacol_t *datacol, index_param_t *index_param, param_t *param, peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table,int indiv_nr);
int put_hkl_to_object_list_from_ems(object_list_t *obj, datacol_t *datacol, index_param_t *index_param, param_t *param, peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table,int indiv_nr);
int put_hkl_to_object_list_from_calc_x_pos (object_list_t *obj, datacol_t *datacol, index_param_t *index_param, param_t *param, peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table, int indiv_nr);
int overlap_check_3D(object_list_t *obj, datacol_t *datacol, index_param_t *index_param, param_t *param, peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table);
int same_hkl(int *G, int *H);
object_list_t *clear_object_list (char *dir, object_list_t *obj);
object_list_t *put_together_object_list(object_list_t *obj, param_t *param, datacol_t *datacol, peaksearch_t*peaksearch);
int put_local_peaks_into_refine_peaks(peaks_t *local_peaks,peaks_t *refine_peaks,int ind_reorient, peaksearch_t *peaksearch,double *om);
int reduce_K_alpha_2(peaks_t *peaks_alle,peaks_t *peaks, param_t *param,datacol_t *datacol);

// index.c
int index_one_reflection(double *h, double *UB, int *H, double *index, double max_dev,double *chisq);
double index_reflections(peaks_t *peaks, double *UB, int additiv, double max_dev);
int least_hkl (peaks_t *peaks, double *UB, double max_laenge, double max_dev, int debug);
int mxinv(double *bx, double *det, double *ax);

// integrate_IPDS.c
void insert_positions_from_fenster(object_list_t *o,int *fenster, az_data_t *data, trans_table_t *trans_table);
void insert_ellipse_geometry(object_list_t *o,peaksearch_t *peaksearch, az_data_t *az_data,az_data_t *az_bg, trans_table_t *trans_table);
int calc_osc_range_from_file_nr(datacol_t *datacol, param_t *param, int run_nr, int file_nr, double *osc_range);
double calc_osc_angle_from_file_nr(datacol_t *datacol, param_t *param, int run_nr, int file_nr);
int calc_file_nr_from_osc_angle(datacol_t *datacol, double osc_angle);
void copy_az_data (az_data_t *data, az_data_t *cp);
int find_az_max_pos (GSList *value,int *pos, az_data_t *data);
void print_plain_peak  (data_t *data, int *pos,int fenster, int log);
void make_az_cross(az_data_t *az_data, int *pos, trans_table_t *trans_table, datacol_t *datacol, data_t *data);
object_list_t *peaksearch_az_plate (az_data_t *az_data, az_data_t *az_bg, object_list_t *object_list, peaksearch_t *peaksearch, datacol_t *datacol, trans_table_t *trans_table,param_t *param, double *bg_sigma);
void print_az_peak  (az_data_t *data, az_data_t *bg, int *pos,int *fenster, peak_t *peak, int p, int log, trans_table_t *trans_table);
void print_orth_peak  (az_data_t *az_data, az_data_t *bg, int *pos, int log, trans_table_t *trans_table);
double interpret_peak (int integral, int untergrund, peak_t *peak, double diff_const) ;
int scan_peak_to_maximum (az_data_t *az_data, az_data_t *az_bg, int *pos,int *fenster, peaksearch_t *peaksearch, trans_table_t *trans_table,param_t *param, datacol_t *datacol);
int compare_values(value_data_t *d1, value_data_t *d2);
GSList *put_az_data_in_value(az_data_t *data, az_data_t *bg,int r_min, GSList *value_list, double *bg_sigma);
void copy_data (data_t *data, data_t *cp);
GSList *new_clear_az_maximum (az_data_t *az_data, az_data_t *data, az_data_t *bg, object_list_t *o, GSList *value);
int find_breite (az_data_t *data, az_data_t *bg, int *pos,int *breite, int *max_pos, param_t *param, int max_beite, int *n_rel);
double find_peak_geometry (object_list_t *o,az_data_t *data, az_data_t *bg, int *pos,param_t *param, int r_min, int r_max,double max_breite, int min_n_rel, trans_table_t *trans_table, double akzeptance);
int find_index_from_phi(int r,double phi,az_data_t *az_data,trans_table_t *trans_table);

int make_az_line( az_data_t *az_data, object_list_t *o);
int integrate_single_object(object_list_t *o, az_data_t *az_data, az_data_t *az_bg, int typ);
int new_integrate_az_peak (object_list_t *o, az_data_t *data, az_data_t *bg, int *pos, peaksearch_t *peaksearch, trans_table_t *trans_table,param_t *param);
object_list_t *new_integrate_az_plate (az_data_t *az_data, az_data_t *az_bg, object_list_t *object_list, peaksearch_t *peaksearch, datacol_t *datacol, trans_table_t *trans_table,param_t *param, double *bg_sigma);

void calc_phi_at_r(int *pos, int r_pos, az_data_t *data, int *fenster, double phi, int *phi_pos);
void strip_standard_box(az_data_t  *az_data, az_data_t *az_bg,int *pos, int *fenster);
void make_short_az_box(az_data_t *data, int *pos, int *fenster, int teiler);
void calc_lorentz_correct (peak_t *peak, datacol_t *datacol);


// background.c
int remove_hotspots (data_t *data, double sigma);
int find_radius(data_t *data);
//double calc_radius_from_pixel (int i,int j, int *origin, double scale, double *correct, double distance);
double *create_bg_sigma(az_data_t *az_data);
void find_az_background(az_data_t *az_data, az_data_t *bg,peaksearch_t *peaksearch,datacol_t *datacol, trans_table_t *trans_table, double *bg_sigma);
int print_az_line_scan(az_data_t *data, az_data_t *bg,int r);

// geometry.c
void put_calibrate (datacol_t *datacol);
int compare_geo_arr(geo_arr_t *d1, geo_arr_t *d2);
void calc_x_from_pos(int *pos, int *origin, double scale, double *x_pos);
void calc_x_from_az_pos(int *pos, az_data_t *data, double *x_pos, trans_table_t *trans_table);
void calc_x_from_rp(int rp, az_data_t *data, double *x_pos, trans_table_t *trans_table);
void calc_pos_from_x(int *pos,  datacol_t *datacol, data_t *data, double *x_pos);
int calc_az_pos_from_x(int *pos, datacol_t *datacol, az_data_t *data, double *x_pos, trans_table_t *trans_table);
double calc_az_distance(int *p1, int *p2, az_data_t *az_data, trans_table_t *trans_table);
void calc_r_p_from_rp(az_data_t *data, int rp, int *rp_pos);
void calc_dir_cos (peak_t *ref, datacol_t *datacol, double *om);
void get_koord_from_line(double abschnitt,double steigung, int x, int y,int *x1,int *x2);
int int_line (data_t *data, double abschnitt, double steigung, double *mittelwert);
void verfeinere_gerade(data_t *data, double *pos1,double *pos2, double *Stei, double *Absch);
void invers_gnomonisch(double *vec, double *p);
int calc_schnittpunkt (data_t *data,gerade_t *gerade, int index1, int index2, int *schnittpunkt);
int check_winkel (double *v1, double *v2);
void pixel_to_proj (double *P,int *S,double scale, int *origin);
void calc_schnittpunkte (gerade_t *gerade, int *index, data_t *data, double *normal);
int compare_gerade(gerade_t *g1, gerade_t *g2);
void pos_to_proj_corr (double pos[2], double *projection,int *anzahl_projections);
void pos_to_proj (double pos[2], double *projection, double min_max[2][2],int *anzahl_projections);
void az_to_view(az_data_t *az_data,data_t *data, trans_table_t *trans_table);
void init_detector(datacol_t *datacol);
void init_az_data( az_data_t *az_data, trans_table_t *trans_table,datacol_t *datacol);
void free_az_data(az_data_t *data);
void free_trans(trans_table_t *trans_table);
void new_interpret_geometry (trans_table_t *trans_table, datacol_t *datacol);

// metrik_IPDS.c
void calc_om(double *om,double *normal,double *laengen);
void reduce_om(double *om);
double fourier_analyse (double *line, int anzahl, double incr, int debug, double min_A,double max_A);
void find_maxima(double *bg,int ncnt, double incr,int **maxima, int *anzahl, int debug);
int vergleiche_metrik(double *om, double *om2);
int check_new_G(double *G, double *G2);
int check_new_om(double *om, double *new_om,double toleranz);
void kreuz_und_quer_produkt(double *normal);
void calc_h_from_x_pos (double *h, double *x_pos, datacol_t *datacol, double osc_angle, double stoe_ii_phi);
void calc_h_from_x_pos2(datacol_t *datacol, double phi, double stoe_ii_phi, double *x_pos, double *h);
void  message_om (double *om, peaks_t *peaks,index_param_t *index_param, char *messgae);
void print_om (double *om, peaks_t *peaks,index_param_t *index_param);
double check_metrik (double *om, loesung_t **loesung, int *anzahl_loesungen, int individuum, peaks_t *peaks, index_param_t *index_param, param_t *param, gnomonisch_t *gnomonisch);
int generate_mehrling( double *achse, int zaehligkeit, datacol_t *datacol, param_t *param);
int check_bewegung(peaks_t *peaks_alle, datacol_t *datacol, index_param_t *index_param, param_t *param,double *winkel, double *euler1, double *euler2, peaksearch_t *peaksearch);
int calc_d_werte (double *A, double *W, peaks_t *peaks, d_wert_liste_t *d_wert_liste, int ordnung, datacol_t *datacol, int lattice_typ);

// view.c
void scale_to_view(data_t *data,data_t *view_data);
void scale_to_range(int minimum, int maximum, data_t *data);
void marView (data_t *data,datacol_t *datacol);

// matrix_integrate_IPDS.c
int compare_valid(valid_peaks_t *v1, valid_peaks_t *v2);
void calc_h_float (int *h_int, double *om, double *h);
int calculate_osc_angle (int *H, double *osc_angle, datacol_t *datacol, double *om, double stoe_ii_phi);
int double_calculate_osc_angle (double *h,  double *osc_angle, datacol_t *datacol, double stoe_ii_phi);
void calc_abgebeugter_strahl (double *h, double *rot, double osc_angle, double stoe_ii_phi, double *primaer, double *s1);
int calc_x_pos_auf_platte (double *h, double osc_angle, double stoe_ii_phi, double *x_pos, datacol_t *datacol);
int check_osc_angle_in_intervall(double *osc_angle, double *intervall);
int check_platte_in_intervall(int *p, int *platten);
calc_peak_list_t *remove_measured_reflections_from_list (peaks_t *peaks, calc_peak_list_t *c_list,datacol_t *datacol, index_param_t *index_param);
void standard_fenster(int *fenster, datacol_t *datacol, peaksearch_t *peaksearch, az_data_t *az_data, double *h);
int check_center( int H[3], int typ);
int is_even(int a);
int is_zero_hkl(int *H);
void calc_platten(int *hkl, datacol_t *datacol,int run_nr,double osc_angle,  param_t *param, peaksearch_t *peaksearch,int *platten);
int pos_is_in_ellipse(int *pos_check,int *pos,int *fenster,az_data_t *az_data);
int pos_is_in_figure(int *pos_check,object_list_t *obj,az_data_t *az_data);
calc_peak_list_t *calc_peaks_one_plate_list (datacol_t *datacol, calc_peak_list_t *calc_peak_list, peaksearch_t *peaksearch,param_t *param, az_data_t *az_data,trans_table_t *trans_table, char *bild_datei, double *om, int indiv_nr);
calc_peak_list_t *calc_peaks_to_measure_list (datacol_t *datacol, calc_peak_list_t *c_peak_list, peaksearch_t *peaksearch,param_t *param, az_data_t *az_data,trans_table_t *trans_table, double *om, int run_nr,int *platten, int clear, int indiv_nr);
int overlapped(object_list_t *cp, object_list_t *cp2, az_data_t *az_data, int tol);
void overlap_check (calc_peak_list_t *c_peak_list, peaksearch_t *peaksearch, az_data_t *az_data,datacol_t *datacol,trans_table_t *trans_table);
int beugungs_bedingung (double *osc_angle, datacol_t *datacol, double EMS, int *H, param_t *param);
object_list_t *check_shadow(object_list_t *obj, position_list_t *position_list);

// xfig.c
FILE *write_xfig_header(char *filename, int n, int *origin, datacol_t *datacol);
void print_peak(FILE *fp, int *pos, double I, double Max);
void print_hkl(FILE *fp, int *pos, int *hkl);
void simul_aufnahme(param_t *param, datacol_t *datacol, peaks_t *peaks_alle, peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table);
void fill_I_from_peaks(peaks_t *peaks1, peaks_t *peaks_alle);
void calc_xfig_pos(datacol_t *datacol, int *pos, int *xfig_pos);

// pulver.c
double generate_powder_scan (az_data_t *data, az_data_t *az_bg, double *powder, datacol_t *datacol, param_t *param, trans_table_t *trans_table, int remove_background);

// correlation_IPDS.c
void calc_max_laenge (metrik_t *metrik, double *l);
void set_diff_laenge(diff_t *diff, int i,double laenge);
void sortiere_ein(int *line,int anzahl_line,double wert, double incr_line);
int find_first_maximum(double *line, int anzahl_line, int mittl, int trigger,int debug, int *fwhm);
double find_max(int *line, int anzahl_line, double incr_line, double *breite, param_t *param);
double calc_correllation (peaks_t *peaks, double *v, datacol_t *datacol, param_t *param);
void refine_laengen(diffs_t *diffs, data_t *data,datacol_t *datacol, index_param_t *index_param, param_t *param);
int compare_diffs(diff_t *d1, diff_t *d2);
int compare_trans(trans_t *d1, trans_t *d2);
void get_maximum_pos (int *P, int *maxima, int anzahl_maxima);
int select_richtung(double *UB, int *index,int anzahl_translationen, trans_t *trans, param_t *param);
int check_UB(double *UB, peaks_t *peaks_alle, index_param_t *index_param, param_t *param);
int find_halbwertsbreite(data_t *data);
int fill_kugel_Daten(data_t *data, double *projection,int anzahl_projections, param_t *param);
int find_maxima_in_Daten (data_t *data, int **maxima, int *anzahl_maxima, param_t *param);
int copy_differenz_vektoren(diffs_t *diffs, diffs_t *alle_diffs,index_param_t *index_param);
int fill_differenz_vektoren (peaks_t *peaks,diffs_t *diffs,datacol_t *datacol, index_param_t *index_param, param_t *param);
int clear_first_maximum(double *line,int anzahl_line, double incr_line, datacol_t *datacol, index_param_t *index_param, param_t *param);

// mehrling.c
int check_mehrling (double *om, int anzahl_translationen, trans_t *trans, int index[3],loesung_t **loesung, int *anzahl_loesungen, peaks_t *peaks_alle, index_param_t *index_param, param_t *param, gnomonisch_t *gnomonisch);

// refine.c
void calc_delta_lattice (double *delta, int typ, double *g,int anzahl_parameter, double *delta_A, double *delta_W);
int calc_phi_chi_psi_from_om (double *om, double *g, double *phi, double *chi, double *psi);
void interpret_om(double *om, double *euler);
void refine_osc_angle_values(peaks_t *peaks, double *ub, datacol_t *datacol);
void print_fehler(double *delta_A, double *delta_W, double *om, char *message);
int do_fix_metrik_search (double *A, double *W, double *om, peaks_t *peaks_alle, datacol_t *datacol,param_t *param, index_param_t *index_param,loesungen_t *loesungen);
void fill_R (double phi, double chi, double psi, double *R);
double refine_orientation (double *ub, peaks_t *peaks, int typ, int max_i, double *delta, int debug, index_param_t *index_param, double *delta_A, double *delta_W, datacol_t *datacol, double *euler);
int take_peaks_from_d_wert_liste(peaks_t *peaks, d_wert_liste_t *d_wert_liste, double tolerance);
int remove_peaks_d_wert_liste(peaks_t *peaks, d_wert_liste_t *d_wert_liste, double tolerance);
void refine_pl_peaks_svd(datacol_t *datacol,pl_peak_list_t *pl_peak_list, index_param_t *index_param,peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table,param_t *param, int max_i, int typ);
void refine_pl_peaks (datacol_t *datacol,pl_peak_list_t *pl_peak_list, index_param_t *index_param,peaksearch_t *peaksearch, az_data_t *az_data, trans_table_t *trans_table,param_t *param, int max_i, int typ);
double index_from_x_pos(pl_peak_list_t *calc_pl, pl_peak_list_t *pl, param_t *param ,double tol);
pl_peak_list_t *insert_calc_pl_list (pl_peak_list_t *pl_peak_list, datacol_t *datacol, peaksearch_t *peaksearch, int run_nr, int indiv_nr);
void index_pl_list (pl_peak_list_t *calc_pl_list, pl_peak_list_t *pl_peak_list, param_t *param, double x_tol);
pl_peak_list_t *free_pl_list(pl_peak_list_t *pl);
void calc_pos_x_and_abl(double *h, double phi, double stoe_ii_phi, double D, double *ro, double *s_1, double *g, double *euler, double *a1, double *a2, double *origin, double *xcalc, int typ);

// loesung.c
void print_mehrling_loesung(peaks_t *peaks,double *om, int individuum, index_param_t *index_param);
void print_mehr_loesungen(loesungen_t *loesungen, peaks_t *peaks_alle, index_param_t *index_param, param_t *param, char *file);
void print_loesungen(loesung_t *loesung, int anzahl_loesungen, FILE *fp,double *om,peaks_t *peaks_alle, index_param_t *index_param);
int compare_loesung(loesung_t *l1, loesung_t *l2);
int project (peaks_t *peaks, data_t *data, int type,param_t *param, index_param_t *index_param, gnomonisch_t *gnomonisch);



// projections.c
int proj_gnomonisch (double x[3], double koord[2]);
int fast_peak_search (data_t *data,peaks_t *peaks, double *normal,  param_t *param);
void diskretisiere_data (data_t *data);
int calc_geraden(data_t *data, geraden_t *gerade, peaks_t *peaks, param_t *param);
void fill_normale(gerade_t *gerade, data_t *data, param_t *param);
int find_dreieck(int index[3],data_t *data,gnomonisch_t *gnomonisch);
double try_triangle (peaks_t *peaks, double *om,index_param_t  *index_param, peaks_t *peaks_alle, param_t *param);
int fill_Daten(data_t *data,double min_max[2][2], double *projection,int anzahl_projectionen, param_t *param);
double check_periodicity (double laenge,double *line,int anzahl_projektions_punkte,double incr, peaks_t *peaks, param_t *param);

// refine_geometry_svd.c
int calc_x_pos_and_ableitungen(double *h,double phi, double stoe_ii_phi, double D,double *origin,double *r_p,double *r_r,
			       double *g,double *euler,double lambda, double *a);
int calc_x_pos_auf_platte2(double *H, double phi, double stoe_ii_phi, datacol_t *datacol, double *x_pos);
int refine_geometry_all_svd (datacol_t *datacol, peaks_t *peaks, double *delta,param_t *param,index_param_t *index_param, int max_i, az_data_t *az_data, double *delta_A, double *delta_W);

// refine_geometry.c
int refine_geometry_all (datacol_t *datacol, peaks_t *peaks, double *delta,param_t *param,index_param_t *index_param, int max_i, az_data_t *az_data, double *delta_A, double *delta_W);
int refine_geometry (datacol_t *datacol, peaks_t *peaks, double *delta,param_t *param,index_param_t *index_param, int max_i, az_data_t *az_data);

// interpret_objects.c
int xds_to_stoe_calibration(datacol_t *datacol);
int stoe_to_xds_calibration(datacol_t *datacol);
object_list_t *interpret_object_list_single (datacol_t *datacol, param_t *param, az_data_t *az_data, object_list_t *obj, index_param_t *index_param, peaksearch_t *peaksearch, trans_table_t *trans_table);
object_list_t *interpret_objects_multi (datacol_t *datacol, param_t *param, az_data_t *az_data, object_list_t *obj, index_param_t *index_param, peaksearch_t *peaksearch, trans_table_t *trans_table);
int interpret_calibration(datacol_t *datacol, param_t *param);

// rec_map.c
int fill_rec_table (rec_h_t *rec_h, data_t *r_map,datacol_t *datacol, rec_map_t *rec_map, index_param_t *index_param, param_t *param, int typ);
int fill_rec_table_crystall (rec_h_t *rec_h, data_t *r_map,datacol_t *datacol, rec_map_t *rec_map, index_param_t *index_param, param_t *param);
//int fill_rec_table (rec_h_t *rec_h, data_t *r_map,datacol_t *datacol, rec_map_t *rec_map, index_param_t *index_param, param_t *param);
int fill_rec_map(data_t *r_map, rec_h_t *rec_h, param_t *param, datacol_t *datacol);

// config_IPDS.c
void read_config_file (view_t *view, rec_map_t *rec_map, param_t *param, peaksearch_t *peaksearch, index_param_t *index_param,gnomonisch_t *gnomonisch,datacol_t *datacol, metrik_t *metric);

// objects.c
object_list_t *remove_object_from_list(object_list_t *tmp,object_list_t *o_list);
object_list_t *append_object_to_list(object_list_t *o);
object_list_t *insert_object_into_list(object_list_t *o,object_list_t *o_pos);
object_list_t *insert_object_at_start (object_list_t *o,int platte,int *pos, int integral, int untergrund, double stoe_ii_phi, int overflow, int nr, int noff);
object_list_t *merge_list_object(object_list_t *o1,object_list_t *o2, object_list_t *start, int *test);
object_list_t *delete_obj_from_list (object_list_t *o2, object_list_t *start);
int remove_all_hkl_from_list(object_list_t *o);
int remove_all_positions_from_list(object_list_t *o);
int add_positions_to_list(object_list_t *o,int r, int *breite2);
geo_list_t *insert_geo_in_list (int r, double phi, int ij, geo_list_t *geo_list);
geo_list_t *insert_geo_at_start (int r, double phi, int ij, geo_list_t *geo_list);
int node_compare(const void *a, const void *b);
int list_len(geo_list_t *l);
geo_list_t** getNodeArr(geo_list_t *l, int *arrLen);
geo_list_t *list_quicksort(geo_list_t *l);
object_list_t *free_obj_list(object_list_t *object_list);
calc_peak_list_t *free_calc_peak_list(calc_peak_list_t *calc_peak_list) ;
calc_peak_list_t *insert_peak_at_start (calc_peak_list_t *cp,int *pos, int *H, int indiv_nr,int *calc_platten);
calc_peak_list_t *remove_measured_reflections_from_list (peaks_t *peaks, calc_peak_list_t *c_list,datacol_t *datacol, index_param_t *index_param);
calc_peak_list_t *remove_calc_peak_from_list(calc_peak_list_t *tmp,calc_peak_list_t *o_list);
peak_list_t *insert_p_at_start (peak_list_t *pl, calc_peak_list_t *o, int file_nr, int run_nr);
int put_fertig_peak_into_peak (peak_t *peak, peak_list_t *obj, trans_table_t *trans_table, datacol_t *datacol, az_data_t *az_data, index_param_t *index_param, param_t *param);
peak_list_t *move_peak_to_fertig (peak_list_t *f_p,peak_list_t *p, peak_list_t **peak_list);
int count_object_list (object_list_t *o);
int count_peak_list (calc_peak_list_t *o);
pl_peak_list_t *insert_pl_peak_at_start (pl_peak_list_t *o,double *x_pos, int platte, double stoe_ii_phi);

//ikosaeder.c
double refine_iko_orientation (double *ub, peaks_t *peaks, int typ, int max_i, double *delta, int debug,index_param_t *index_param, double *delta_A, double *delta_W, datacol_t *datacol, double *euler);
int iko_to_cart(double *iko, double *ub);
double get_iko_lattice_from_OM (double *ub);
int fill_W_mat_from_iko(double a, double *W);
int cart_to_iko(double *ub, double *iko);
int ev_to_ec(double *ev,double *ec);
int ec_to_ev(double *ec,double *ev);

//refine_twin.c
int do_twin_refinement(datacol_t *datacol, param_t *param, index_param_t *index_param, peaksearch_t *peakserach, az_data_t *az_data, trans_table_t *trans_table, int do_twin, int anzahl, double *twin_op);
double refine_twin_orientation (pl_peak_list_t *pl, datacol_t *datacol, int typ, int max_i, int debug,index_param_t *index_param, int do_twin, int anzahl,double *twin_op);

#endif
