#ifndef BILD_H
#define BILD_H
#define OVERFLOW_PIXEL 256001

#define OK (0)

#ifndef WARN
#define WARN "# [31;1mWARNING[0m"
#endif
#ifndef ERR
#define ERR "# [31;1mERROR[0m"
#endif
#ifndef DEG_RAD
#define DEG_RAD 57.29577951
#endif

#ifndef max
#define max(a,b) ( (a)>(b)? (a):(b) )
#endif
#ifndef min
#define min(a,b) ( (a)<(b)? (a):(b) )
#endif
#include<stdio.h>
#include<unistd.h>
#include<stdlib.h>
#include<assert.h>
#include<strings.h>
#include <string.h>
#include<tiffio.h>
#include<getopt.h>
#include <errno.h>
//#include"moi2pck.h"
#include<fftw.h>
#include<rfftw.h>
#include<math.h>
#include<malloc.h>
#include<png.h>
#include<pdflib.h>
#include <cpgplot.h>

#include "../libcryst/cryst.h"

#define STOE_I 1
#define STOE_II 2
#define MAR180 3
#define MAR345 4
#define MAR_CCD 5
#define BNL 6
#define BRUKER 7
#define ESRF 8
#define TIFF_FORMAT 9
#define RAW 10
#define PNG 11
#define MAR300 12

typedef struct {
  int teiler;
  unsigned int *Daten;
  int x;
  int y;
  int origin[2];
  int maximum;
  int gesamtzahl;
  int overflows;
  double scale;
  double view_scale;
  double mittelwert;
  double stoe_ii_phi;
  double osc_angle;
  double mittl_untergrund;
} data_t;

typedef struct {
  int overflows;
  double drehwinkel;
  double stoe_ii_phi;
  double distance;
  double drehintervall;
  double lambda;
  double bel_zeit;
  double kV;
  double mA;
  double temperatur;
  int minimum;
  int maximum;
  double mittelwert;
  char name[256];
} bild_header_t;

typedef struct {
  char program[256];
  char date[256];
  char format[256];
  int high;
  int pixelsize[2];
  int roff;
  int toff;
  int gaps[8];
  int adc_a;
  int adc_b;
  int adc_add_a;
  int adc_add_b;
  double multiplyer;
  double gain;
  double lambda;
  double distance;
  double resolution;
  double phi_start;
  double phi_end;
  double delta_phi;
  double omega_start;
  double omega_end;
  double delta_omega;
  double chi;
  double twotheta;
  double center[2];
  char mode[256];
  double time;
  int count_start;
  int count_end;
  int count_nmeas;
  double MinCnt;
  double MaxCnt;
  double AvrCnt;
  double SIG;
  int min_I;
  int max_I;
  double average_I;
  double sigma_I;
  int hist_start;
  int hist_end;
  int hist_max;
  double kV;
  double mA;
  char monochromator[256];
  double collimator[2];
  double slit_1[2];
  double slit_2[2];
  double beamstop;
  double xtal;
  double trans[2];
  double rot[2];
} marheader_t;

typedef struct {
    int total_pixels_x;
    int total_pixels_y;
    int lrecl;
    int max_rec;
    int overflow_pixels;
    int overflow_records;
    int counts_per_sec_start;
    int counts_per_sec_end;
    int exposure_time_sec;
    int programmed_exp_time_units;
    float programmed_exposure_time;
    float r_min;
    float r_max;
    float p_r;
    float p_l;
    float p_x;
    float p_y;
    float centre_x;
    float centre_y;
    float lambda;
    float distance;
    float osc_angle_start;
    float osc_angle_end;
    float omega;
    float multiplier;
    char scanning_date_time[24];
    char filler[ (600 - 31)*sizeof(int) ];
} marimageheader;

int get_image_type (char *filein);

int tomar (char *filename, short *mar, int packflag, marimageheader h);
void create_file_name (char *fileroot, char *filein, int type, char *filenname, 
		       int otype);
int unpack_word(FILE *packfile, int x, int y, short int *img);


int read_raw (char *filein, unsigned int **Daten, int *x, int *y);
int read_tiff (char *filein, unsigned int **Daten, int otype, int *x, int *y);
int read_edf (char *filein, unsigned int **Daten, int *bits, int *x, int *y);
int read_tra (char *filein, unsigned int **Daten, int *x, int *y);
int read_tra_header  (char *filein, bild_header_t *bild_header);
int read_xi_header  (char *filein, bild_header_t *bild_header);
int read_xi (char *filein, data_t *data, datacol_t *datacol);
int read_short_tra (char *filein, unsigned int **Daten, int *x, int *y);
int read_bnl(char *filein, int bin_size, unsigned int **Daten, int *x, int *y,
	     double *distance, double *lambda, double *osc_start, double *osc_range,
	     double *time);
int read_saint (char *filein, unsigned int **Daten, int *x, int *y);
int binning (unsigned int **Daten, int bin_size, int *x, int *y, int MAX);
int short_binning (unsigned short **Daten, int bin_size, int *xx, int *yy);
void find_maximum (unsigned int *Daten, int x,int y, int *MAX, int *MIN);
void find_short_maximum (unsigned int *Daten, int x,int y, int *MAX, int *MIN) ;
void scale_up (unsigned int *Daten,int x,int y);
void scale_up_short (unsigned int *Daten,int x,int y, int wert);
void make_cross(unsigned int *Daten,int x, int y, int pos1, int pos2);
void make_short_cross(unsigned int *Daten,int x, int y, int pos1, int pos2, int teiler);
void make_short_box(unsigned int *Daten,int x, int y, int *pos, int *fenster, int teiler);
void mittle_Daten (unsigned int *Daten, int x, int y, int breite);
void view (char *fileroot,unsigned int *Daten, int x, int y, int teiler);
int find_background (unsigned int *Daten);
int to_short (unsigned int *Daten, unsigned short **Daten_ushort, int size, int MAX);
void invert_data(unsigned int *Daten, int size, int MAX);
void logarith_data (unsigned int *Daten, int size, int log, int MIN, int MAX);
int fit_to_new_dimension (unsigned int **Daten, int *x, int *y, int x_neu, int y_neu);
int fit_to_new_dimension_short (unsigned short **Daten, int *x, int *y, int x_neu, int y_neu);
void edf16_write  (char fname[80],unsigned  short *Daten_ushort, int x, int y);
void edf32_write  (char fname[80], unsigned int *Daten, int x, int y);
void write_zimg (char *filename, unsigned int *Daten, int size);
void write_short_zimg (char *filename, unsigned short *Daten, int size);
void write_colormap (char *filename);
int write_pdf (char fname[80], unsigned int *Daten, int x, int y);
int write_png (char fname[80], unsigned int *Daten_ushort, int x, int y);
int write_ps (char fname[80], unsigned int *Daten, int x, int y, int MIN, int MAX);

int new_read_pck (char *filein, unsigned int **Daten, int *x, int *y);
int read_mar_header(char *filein, bild_header_t *bild_header);
int read_pck (char *filein, unsigned int **Daten, int *x, int *y);
int read_png(char *file_name,unsigned int **Daten, int *x ,int *y);
void falte_gauss(unsigned int **Daten, int *x, int *y, int sigmax, int sigmay);
void falte_short_gauss(unsigned short **Daten, int *x, int *y, int sigmax, int sigmay);
void init_gauss (fftw_complex **gausst, int x, int y,float sigmax, float sigmay);
void fill_gauss (fftw_real *gauss, int x, int y,float sigmax, float sigmay);
int read_pck (char *filein, unsigned int **Daten, int *x, int *y) ;
int read_tiff (char *filein, unsigned int **Daten, int otype, int *x, int *y) ;
int read_edf (char *filein, unsigned int **Daten, int *bits, int *x, int *y);
void edf16_write  (char fname[80], unsigned short *Daten_ushort, int x, int );
void edf32_write  (char fname[80], unsigned int *Daten, int x, int y);
int read_saint_header (FILE *fp, int *x, int *y);
int read_saint (char *filein, unsigned int **Daten, int *x, int *y);
int read_tra (char *filein, unsigned int **Daten, int *x, int *y);
int read_short_tra (char *filein, unsigned int **Daten, int *x, int *y);

#define PACKIDENTIFIER "\nCCP4 packed image, X: %04d, Y: %04d\n"
#define PACKBUFSIZ BUFSIZ
#define DIFFBUFSIZ 16384L
#define shift_left(x, n)  (((x) & setbits[32 - (n)]) << (n))
#define shift_right(x, n) (((x) >> (n)) & setbits[32 - (n)])
void read_marheader(marheader_t *marheader, char *filein);
int packi32(int i32, short int i16);
void	gethead	(int *mar345, FILE *image, int *x, int *byteswap, int *kov);
void	getpck		(short int *, int strong[2], short int i16, int i32, char fff[128],int mar345, FILE *image,int x, int y, int kov, int byteswap, int cutoff);
void int_getpck(int *img, int strong[2],short int i16, int i32, char fff[128], int mar345, FILE *image,int x, int y, int kov, int byteswap, int cutoff);

int read_mar_ccd_header (char *filein, bild_header_t *bild_header);
int read_mar_ccd (char *filein, data_t *data, datacol_t *datacol);

#endif
