#ifndef CRYST_H
#define CRYST_H

#include<stdio.h>
#include<math.h>
#include <malloc.h> 
#include<assert.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include<fftw.h>
#include<rfftw.h>
#include<string.h>

#ifndef WARN
#define WARN "# [31;1mWARNING[0m"
#endif
#ifndef ERR
#define ERR "# [31;1mERROR[0m"
#endif
#ifndef GUT
#define GUT "# [32;1mGUT[0m"
#endif
#ifndef DEG_RAD
#define DEG_RAD 57.295779513082322864647721872
#endif

#ifndef ROT
#define ROT "[31;1m"
#endif
#ifndef GRUEN
#define GRUEN "[32;1m"
#endif
#ifndef ENDE
#define ENDE "[0m"
#endif

//#ifndef NOSYSTEM
//#define NOSYSTEM 1
//#endif
//#undef NOSYSTEM

#ifndef max
#define max(a,b) ( (a)>(b)? (a):(b) )
#endif
#ifndef min
#define min(a,b) ( (a)<(b)? (a):(b) )
#endif

#ifndef sqr
#define sqr(x) (x)*(x)
#endif

#ifndef SIGN
#define SIGN(a,b) ((b)>=0.0 ? fabs (a): -fabs(a))
#endif
#define NO_METRIK -1
#define TRIKLIN 0
#define MONOKLIN 1
#define ORTHORHOMBISCH 2
#define TETRAGONAL 3
#define KUBISCH 4
#define TRIGONAL 5
#define HEXAGONAL 6
#define OKTAGONAL 8
#define DEKAGONAL 10
#define IKOSAEDRISCH 20

#define PRIMITIV 0
#define I_ZENTRIERT 1
#define F_ZENTRIERT 2
#define A_ZENTRIERT 3
#define B_ZENTRIERT 4
#define C_ZENTRIERT 5
#define R_ZENTRIERT 6


#define A_ACHSE 0
#define B_ACHSE 1
#define C_ACHSE 2

#define LABOR_KOORDINATEN 0
#define KRISTALL_KOORDINATEN 1

#define COS15 0.96592583
#define SIN15 0.25881905
#define COS30 0.866025403
#define SIN30 0.5

#define XDS_CORRECT 1
#define STOE_CORRECT 2

typedef struct {
  int anzahl;
  int *H;
  double *h_laenge;
} d_wert_liste_t;

typedef struct {
  int correction_typ;
  double stoe_spiral_offset;
  double correct[8];
  double rot_achse[3];
  double rot_achse_unkorrigiert[3];
  double rot_rot[2];
  double primaerstrahl[3];
  double primaerstrahl_unkorrigiert[3];
  double rot_prim[2];
  double origin[2];
  double origin_correct[2];
  double delta[17];
} calibration_t;

#define MONOCHROMATOR TRUE
#define SYNCHROTRON FALSE

typedef struct {
  int detector_typ;
  double detector_radius;
  calibration_t calibration[1];

  int erste_platte;
  int anzahl_runs;
  int run_nr;
  int anzahl_bild_dateien;
  int platten[2];
  int pol_typ;
  int binning;
  int test_H[3];

  double scaling_faktor;

  double polarization;
  double polarization_normal[3];
  double monochromator_angle;
  double distance;
  double lambda;
  double osc_angle;
  double osc_angle_start;
  double osc_angle_wert;
  double stoe_ii_phi;
  double delta_osc_angle;
  double K_alpha[2];
  double delta_h;
  double time;
  double diff_const;

  double om[9];
  double sicher_om[9];
  double delta_lattice[6];
  int anzahl_oms;
  double *nom;
  double *delta_nom;
} datacol_t;

typedef struct {
  double phi;
  double chi;
  double omega;
  double theta;
  double ny;
  double psi;
} diff6_winkel_t;

typedef struct {
  int H[3];
  int H_iko[6];
  int H_deca[5];
  int remove;
  int indexed;
  double h[3];
  double I; 
  double sigma_F2;
  double P;
  double Lorentz;
  double laenge;
  //  double calc_laenge;
  //  double delta_laenge;
  double x_pos_calc[2];
  double dir_cos[6];  
} ref_t;

typedef struct {
  ref_t ref;
  int pos[2];
  int fenster[4];
  //int anzahl_platten;
  int platten[2];
  int integral;
  int untergrund;
  int indiv_nr;
  int overflow;
  int boundary;
  int overlap;
  double sort;
  double sr;
  double osc_angle;
  double stoe_ii_phi;
  double x_pos[2];
} peak_t;

typedef struct {
  int pos[2];
  int fenster[4];
  int integral;
  int untergrund;
  int platte;
  int anzahl_hkl;
  int overflow;
  int boudary;
  double osc_angle;
  double stoe_ii_phi;
  int *indiv_nr;
  int *nhkl;
  int run_nr;
  int run_offset;
} object_t;

typedef struct {
  int anzahl;
  int anzahl_indiziert;
  peak_t *peak;
  double scale;
  double laengster_vektor;
} peaks_t;

typedef struct pe_list {
  int H[3];
  int integral;
  int untergrund;
  int platten[3];
  int pos[2];
  int overflow;
  int boundary;
  int overlap;
  int file_nr;
  int run_nr;
  int indiv_nr;
  double stoe_ii_phi;
  struct pe_list *next;
  struct pe_list *previous;
} peak_list_t;

typedef struct c_list {
  int calc_platten[2];
  int pos[2];
  int H[3];
  int integral;
  int untergrund;
  int shadowed;
  int indiv_nr;
  int overflow;
  int boundary;
  int overlap;
  double stoe_ii_phi;
  struct c_list *next;
  struct c_list *previous;
} calc_peak_list_t;

typedef struct {
  double t[9];
  double a[3];
  double w[3];
  double g[9];
  double volumen;
  char klasse[2];
  char center[2];
  int out_of_range;
  int index;
} delos_t; 

typedef struct {
  int H[3];
  int indexed;
  int remove;
  int anzahl_aufnahmen;
  double P;
  double Lorentz;
  double osc_angle;
  double x_pos[2];
  double h[3];
  double laenge;
  double calc_laenge;
  double delta_laenge;
  double delta_h;
  double F;
  double sigma_F2;
  double I;
  //  double sigma_I;
  double dir_cos[6];
} REF_t;

typedef struct {
  int anzahl;
  double laengster_vektor;
  REF_t *ref;
} refl_t;

typedef struct {
  double R[9];
} matrix_t;

typedef struct {
  int nsym;
  char label[12];
  matrix_t matrix[24];
} laue_t ;

// cryst.c:
int rotate_b_direct(double *om, double *om2);
double calc_theta (double v[3], double lambda);
double calc_laenge_metrik (double *a, double *G);
double  calc_laenge (double *v);
void OM_to_G (double *UB,double *G_inv);
void B_to_G (double *B,double *G);
void calc_Hmax (double *OM, double lambda, int *hmax,double *size,double distance);
int check_metrik_typ(double *A, double *W);
int fill_metrik_matrix_from_angles (double A[3], double W[3], double *g, int debug, int *anzahl_parameter, int typ);
void fill_metrik_tensor_from_angles (double A[3], double W[3], double *g);
void  triklin_metrik_to_OM (double *A, double *W, double *UB);
void  metrik_to_OM (double *g_inv, double *UB);
double get_volume_from_OM (double *om);
void get_lattice_from_OM (double *OM, double *A,double *W);
void get_rec_lattice_from_OM (double *OM, double *A,double *W);
double calc_lambda(double *s, double *h);
void rotate_UB (double phi, double chi, double *UB);
void calc_h_vec (double *UB, double phi, double chi, int H[3], double h[3]);
double calc_max_theta(double size, double distance);
double  calc_aufl_from_theta (double lambda, double theta);
double calc_theta_from_aufl (double lambda, double aufl);
double calc_aufl_from_hkl (int *H, double *OM);
double calc_metrik_laenge (double *vec, double *G_inv);
void direct_to_reciproc (double *d, double *r);
void reciproc_to_direct (double *r, double *d);

// math.c
double sind(double s);
double cosd(double s);
double tand(double d);
double asind(double s);
double acosd(double s);
double atand(double s);
double atand2(double s, double t);
double sign(double x);
void polint(double *xa, double *ya, int n, double *y, double *dy);

// matrix.c
void einheits_matrix(double *O);
void mat_vec(double *UA_0, double *h_float, double *S1);
void mat_vec_N(double *UA_0, double *h_float, double *S1, int N);
void mat_vec_2x2(double *UA_0, double *h_float, double *S1);
void vec_mat( double *h_float,double *UA_0, double *S1);
void copy_matrix (double *m1, double *m2);
void copy_delta (double *m1, double *m2);
void invert_matrix2(double *ax, double *bx);
void invert_matrix(double *a1, double *a_inv);
void invert_matrix_n(double *a1, double *a_inv, int N);
void invert_matrix_2x2(double *a1, double *a_inv);
void transpose (double *M, double *M_T);
void mat_mat (double *m1, double *m2, double *m3);
void  mat_mat_N (double *m1, double *m2, double *m3,int N);
void mat_mat_2x2 (double *m1, double *m2, double *m3);
void matrix_put (double *M);
void lmatrix_put (char *label,double *M);
void matrix_put_n (char *label, double *M, int N);
void matrix_put_2x2 (double *M);
double calc_det (double *M);
double calc_spur (double *M);
void gp_matrix_put (double *M,int lt, char *file);
void shipley_invert(double *A, int N, double *A_inv);

// vector.c
void vector_copy (double *vec, double *erg);
void vector_multiply (double *vec, double A);
void vector_add (double *vec, double *vec2, double *erg);
void normiere (double *v);
void vector_put (char *string, double *M);
void vector_put_n (char *string, double *M, int N);
double skal_prod (double *a, double *b);
double calc_winkel_metrik (double *a, double *b, double *G);
double calc_winkel (double *a, double *b);
void  kreuz_prod (double *a, double *b, double *c);
void rotate_about_achse(double *vec,double *achse,double winkel, double *erg);
void set_laenge (double *vec,double laenge);

// diff.c
void d6_rotation (double vec[3], diff6_winkel_t *diff6_winkel);
void diff6_rotation (double vec[3], double phi,double chi, double ome, double psi);
void rotate_about_chi_omega (double vec[3],double *r_r, double *erg);
void rotate_about_psi_omega (double vec[3], double *r_p, double *erg);
void rotate_about_x (double *vec, double chi, double *erg);
void xyz_rotation (double vec[3], double x,double y, double z);
void rotate_stoe_ii_phi(double v[3],double phi, double *erg);
void rotate_stoe(double v[3],double phi);
double find_best_phi(double phi, double phi1, double phi2);

//gauss.c
void falte_gauss_1D (double *function, int anzahl, double sigma);
void gauss_1D(int size, double sigma, double  **gauss);
void fft_convolute(double *y, double *gauss, int np);
void calc_frequencies (double *fourier, double *frequenzen, int anzahl);
void forward_fourier_1D (double *function, double *ergebnis, int anzahl);
void backward_fourier_1D (double *function, double *ergebnis, int anzahl);

//utils.c
double second(void);
void string_subst(char *ausgang, char *bereich, char *ersetzung, char *ziel, int recursive);

//intensity.c
void gauss_1D(int size, double sigma, double  **gauss);
void bgfit(int np, double* bg, int fwhm, int niter, double sigma);
void background_correct (double *function,int anzahl, int fwhm, double  sigma);
void calc_background (double *function,int anzahl, int fwhm, double  sigma);
void integrate_profile(double *cnt, int ncnt, int fwhm, double bgsigma, double *iobs, double *sigma);
void bgfit(int np, double* bg, int fwhm, int niter, double sigma);
void calc_background_circle (double *function,int anzahl, int fwhm, double ratio);
void bgfit_circle(int np, double* bg, int fwhm, double ratio, int niter);
//void int_bgfit_circle(int np, unsigned int *bg, int fwhm, double ratio, int niter);
//void int_fast_bgfit_circle(int np, unsigned int *bg, int fwhm, double ratio, int niter, int *wurzel);
double calc_bg_sigma(int np, unsigned int *bg, int print, int r);
void bgfit_circle_sigma(int np, unsigned int *bg, int fwhm, double ratio, int niter, double sigma);
void bgfit_linear (int np, double* bg, int fwhm, double ratio, int niter, int i1, int i2);
void mittle (double *f, double *m, int np, int n);

//gerade.c
void regression (double *X, double *Y, int dim,double A[2],double B[2],double *r);

//delaunay.c
double Sign(double a);
void delos(double *A, double *W, int red, double *RED, delos_t *delos);

// transformationen.c
void basis_trafo (double *B1, double *v1, double *B2, double *v2);
void karthesische_koordinaten (double r, double phi, double lambda, double *v);
void kugel_koordinaten (double *v, double *r, double *phi, double *lambda);
void azimutal_koordinaten (double x, double y, double *r, double *phi);
void basistrafo(peaks_t *peaks, double *trafo, double *normal);

// laue.c
int find_transformation_matrix(double *om,double *trans, int lauegrp);
int check_unique(double *hkl, int lauegrp);
int find_unique (int *H, double *h, int *H_new, double *h_new, double *om, int l);
int find_trans (double *h1, double *h2, int l, double *R);
int make_laue_equiv (double *H_int, double *h, double *om, peaks_t *peaks, int l);
void fill_laue (laue_t *laue);
void append_laue_equiv (peaks_t *refl_orig, double *om, int l);
// utils.c
double second() ;

//SVD.c
void svdcmp(double **a, int m, int n, double w[], double **v);
void lfit(double x[], double y[], double sig[], int ndat, double a[], int ia[],
	  int ma, double **covar, double *chisq, void (*funcs)(double, double [], int));
double pythag(double a, double b);
void svdfit(double x[], double y[], double sig[], int ndata, double a[], int ma,
	    double **u, double **v, double w[], double *chisq,
	    void (*funcs)(double, double [], int));
void svbksb(double **u, double w[], double **v, int m, int n, double b[], double x[]);
void covsrt(double **covar, int ma, int ia[], int mfit);
void gaussj(double **a, int n, double **b, int m);

void shipley_invert(double *A, int N, double *A_inv);

// files.c
void write_shelx_hkl(char *filename, peaks_t *peaks);
void read_shelx_hkl(char *filename, peaks_t *peaks );
int read_h_file (char *fileroot,peaks_t *peaks);

// geometry.c
void calc_om_from_dir_cos(double lambda,peaks_t *peaks, double *om);
int calc_diff6_winkel_bysecting(double *v, diff6_winkel_t *diff6, double lambda);
int check_new_laenge(d_wert_liste_t *d_wert_liste, double *h);

#endif
